/* ---------------------------------------------------------------------------------------------------------------------------------------------------------------
// aPubSub system by Toni Wagner  (@itsatony) --------------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------
// ---- this code is part of the VisualWeb Project by ViSERiON UG (haftungsbeschraenkt) (http://www.viserion.com)  ------------------------
// ---- if released, we license it under MIT license. It includes work from others (e.g. utils) - their licenses apply for their parts ---------
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------
//	DEPENDS ON:
//		utils parts of vwc.utils ( general, huge thanks for the cool code and brilliant framework! )
//	BASED ON:
//		various sources .. the pubsub idea ... 
//	TESTS: 
//		use make in this module's root dir
//	VERSION:
//		v0.4 	15.09.11
//	URLs:
//		blogpost: http://coffeelog.itsatony.com/?p=187
//		github: https://github.com/itsatony/vws.pubsub
// --------------------------------------------------------------------------------------------------------------------------------------------------------------- */

var utils = require('./vws.utils');

//var utils = uCol.newCollection();

module.exports = aPubSub;

/* --------------------------------------------------------------- the basic instance of aPubSub  ------------------------------------------------------------------
// description: 
//		PubSub systems are great and widely used today. However, we wanted something that emulates the event-bubbling architecture of DOM events.
//		Next to publications being able to bubble, subscribers can also choose whether they would like to receive bubbled publications.
//		Furthermore, the system has the ability to LOG publications and retrieve them later. This comes in handy when using asynchronously loaded code
// 		or instantiation that needs to refer to old publications. 
//		We created this little thing for use in our framework project. We use it on the serverside with node.js and on the client as jQuery plugin.
//		We will release and update both versions for community use and happy forking. The code is certainly not brilliant or crazy fast, but we tried to
//		have it well-annotated and easy to understand (and extend).
//		We're very happy to learn about ideas, improvements, bugs (and fixes!) .. to get in contact with us, please use twitter ( @itsatony )
//		
//	returns: Object
//		aPubSub ... 
//
//	example:
//			aPubSub.publish('/politics/europe', { content: 'this is wonderfool',  { bubble:true, persist:true }, 'myTestPublisher');
*/	

	function aPubSub() {
		var myScope = this;
		this.defaults  = {
			bubble: true,
			persist: false,
			getBubbles: true,
			scope: myScope,
			debugging: 0			// 0 = none,  1 = console.log all pubs and subs by topic and publisher id
		}	
		this.publicationChannels = {}
		this.publicationLog = new Array();
		this.publicationAnnouncements = new Array();
		return this;
	}

	/* --------------------------------------------------------------- aPubSub.publish  ----------------------------------------------------------
	// description: 
	//		Publish some data on a named topic. 
	// topic: String
	//		The channel to publish on. if no / are given, 'root' will be assumed.
	// data: Array
	//		The data to publish. Each array item is converted into an ordered
	//		argument on the subscribed functions. 
	//		a single attribute is added to the given data: originalTopic  (it stores the complete topic, giving access to the full Topic for Subscribers of lower (bubbled) branches)
	// settings: Object { 
	//			scope: (object) myScope 
	// 						defaults to myScope, if you want a different scope in the callbacks, just supply it here
	//			persist: (boolean) false,   
	// 						defaults to false, if true the publication will be saved in the this.publicationLog object (with a timestamp attached)
	//			bubble: (boolean) false	
	//  					defaults to false, if true a topic will bubble up it's branch: 
	//							/politics/europe/internet/myOpinion  would activate callbacks for 
	//								first '/politics/europe/internet/myOpinion',
	//								then '/politics/europe/internet',
	//								then '/politics/europe', 
	//								then '/politics',
	//								and '/'
	//		}
	//	returns: Object
	//		{ 
	//			'topic':topic, 
	//			'data': data, 
	//			'settings': settings, 
	//			'timestamp': new Date().getTime(), 
	//			'publisher': publisher 
	//		}
	//	example:
	//			aPubSub.publish('/politics/europe', { content: 'this is wonderfool',  { bubble:true, persist:true }, 'myTestPublisher');
	*/	
	aPubSub.prototype.publish = function(topic, data, settings, publisher) {
		var config = utils.extend({ scope: this.defaults.scope, persist: this.defaults.persist, bubble: this.defaults.bubble	}, settings);
		
		topic = this.stripSlashes(topic);
		data.originalTopic = topic;
		if (this.defaults.debugging == 1) {
			console.log('vws.pubsub debugging : [publication] '+topic+' bubbling='+config.bubble+' persistent='+config.persist+' publisher='+publisher);
		}
		var topicTree = [ topic ];
		if (config.bubble) {
			var BranchPoint = topic.lastIndexOf('/');
			while (BranchPoint>-1) {
				var shortenedTopic = topic.substr(0,BranchPoint);
				if (shortenedTopic != '') topicTree.push(shortenedTopic);
				BranchPoint = shortenedTopic.lastIndexOf('/');
			}			
		}
		topicTree.push('/'); //allow subscriptions to root
		for (var branch in topicTree) {
			for (var receiver in this.publicationChannels[topicTree[branch]]) {
				if (branch != 0 && this.publicationChannels[topicTree[branch]][receiver].getBubbles == false) {
					continue; 
				} else {
					this.publicationChannels[topicTree[branch]][receiver].callback.apply(config.scope, [data, topicTree[branch], publisher ]);
					if (this.defaults.debugging == 1) {
						console.log('vws.pubsub debugging : [subscription] '+topicTree[branch]+' getBubbles='+this.publicationChannels[topicTree[branch]][receiver].getBubbles+' subscriber='+receiver);
					}
				}
			}
		}
		var publication = { 'topic':topic, 'data': data, 'settings': settings, 'timestamp': new Date().getTime(), 'publisher': publisher };
		if (config.persist) this.publicationLog.push(publication);
		return publication;
	}

	/* --------------------------------------------------------------- aPubSub.subscribe  ----------------------------------------------------------
	// description: 
	//		Subscribe to publication on certain topics/channels. 
	// topic: String
	//		The channel to listen to. if no / are given, 'root' will be assumed.
	// callback: function
	//		a definition of what to do with the data passed on by the publication
	// getBubbles: boolean 
	//		should the callback also be executed for all publications that were sent via subbranches of the given topic?
	//			example:   topic = '/news/europe'    
	//							with getBubbles=true the callback would be executed for 
	//							'/news/europe', 
	//							but also for '/news/europe/sweden', '/news/europe/germany/bavaria'. etc...
	//	subscriberId: String	
	//			the name/id of the subscriber (potentially for later reference, ... )
	//	returns: Object
	//		{
	//			'topic': topic, 
	//			'id': subscriberId 
	//		}
	//
	//	example:
	//			var mySubscription = aPubSub.subscribe('/politics/europe', function(data, currentBranch, publisher) { console.log(data.originalTopic); },  true, 'myLittleListener');
	*/ 
	aPubSub.prototype.subscribe = function(topic, callback, getBubbles, subscriberId) {		
		if (utils.isFunction(callback) == false) callback = function() { return true; };
		if (typeof getBubbles != 'boolean') getBubbles = this.defaults.getBubbles;
		if (topic == '') topic = '/';
		if (topic != '/') topic = this.stripSlashes(topic);
		if (typeof this.publicationChannels[topic] == 'undefined') {
			this.publicationChannels[topic] = new Array();
		}
		if (typeof subscriberId == 'undefined') subscriberId='subscriber_'+Math.floor(Math.random()*1000000);
		while (typeof this.publicationChannels[topic][subscriberId] != 'undefined') subscriberId = 'subscriber_'+Math.floor(Math.random()*10000);
		this.publicationChannels[topic][subscriberId] = { 'callback': callback, 'getBubbles': getBubbles, 'id': subscriberId }
		return {'topic': topic, 'id': subscriberId }; 
	}

	/* --------------------------------------------------------------- aPubSub.unsubscribe  ----------------------------------------------------------
	// description: 
	//		Remove a subscription. 
	// topic: String
	//		The channel to unsubscribe from.
	// subscriberId: String
	//		the id of the subscriber that should be removed.
	// 
	//	retuns: boolean	
	//			was the subscription found and successfully removed?
	//
	//	example:
	//			var mySubscription = aPubSub.subscribe('/politics/europe', function(data) { console.log(data.originalTopic); },  true, 'myLittleListener');
	//			aPubSub.unsubscribe('/politics/europe', mySubscription.id);
	*/
	aPubSub.prototype.unsubscribe = function(topic, subscriberId) {
		if (typeof this.publicationChannels[topic] != 'undefined' && typeof this.publicationChannels[topic][subscriberId]) {
			delete this.publicationChannels[topic][subscriberId];
			return true;
		}
		return false;
	}

	/* --------------------------------------------------------------- aPubSub.announce  ----------------------------------------------------------
	// description: 
	//		Announce what you will (possible) publish at some point.
	//		this is just for 'cosmetic' purposes and maybe debugging. it simply allows modules to announce their publications and stores these announcements.
	//		hence, if you want to add another module to a project you don't need to read the source to know what's being published, but just check the announcements array.
	// topic: String
	//		The channel to publish on. if no / are given, 'root' will be assumed.
	// data: Array
	//		The data to publish. Each array item is converted into an ordered
	//		argument on the subscribed functions. 
	//		a single attribute is added to the given data: originalTopic  (it stores the complete topic, giving access to the full Topic for Subscribers of lower (bubbled) branches)
	// settings: Object { 
	//			scope: (object) myScope 
	// 						defaults to myScope, if you want a different scope in the callbacks, just supply it here
	//			persist: (boolean) false,   
	// 						defaults to false, if true the publication will be saved in the this.publicationLog object (with a timestamp attached)
	//			bubble: (boolean) false	
	//  					defaults to false, if true a topic will bubble up it's branch: 
	//							/politics/europe/internet/myOpinion  would activate callbacks for 
	//								first '/politics/europe/internet/myOpinion',
	//								then '/politics/europe/internet',
	//								then '/politics/europe', 
	//								then '/politics',
	//								and '/'
	//		}
	//	returns: Object
	//		{ 
	//			'topic':topic, 
	//			'data': data, 
	//			'settings': settings, 
	//			'timestamp': new Date().getTime(), 
	//			'publisher': publisher 
	//		}
	//	example:
	//			aPubSub.announce('/politics/europe', 'some content',  { bubble:true, persist:true }, 'myTestPublisher');
	*/	
	aPubSub.prototype.announce = function(topic, data, settings, publisher) {
		var myAnnouncement = {
			'topic':topic,
			'data':data, 
			'settings':settings,
			'timestamp': new Date().getTime(), 
			'publisher': publisher
		}
		if (typeof this.publicationAnnouncements[topic] == 'undefined') {
			this.publicationAnnouncements[topic] = new Array();
		}
		this.publicationAnnouncements[topic].push(myAnnouncement);
		return myAnnouncement;
	}

	/* --------------------------------------------------------------- aPubSub.wasPublishedOnTopic  ----------------------------------------------------------
	// description: 
	//		sometimes one might want to know whether publications have been sent out in the past (asynchronously loaded script parts, newly instanced stuff, etc.)
	//		if publications were sent with the 'persist:true' setting, they are saved in this.publicationLog .
	//		wasPublishedOnTopic goes through this log and returns an array of publications matching the given topic.
	// myTopic: String
	//		The topic to look for in the publicationLog.
	// getBubbles: boolean
	// 		should also publications on subtopics be returned?
	//
	//	returns: Array
	//		[{ 
	//			'topic':topic, 
	//			'data': data, 
	//			'settings': settings, 
	//			'timestamp': new Date().getTime(), 
	//			'publisher': publisher 
	//		}]
	//	example:
	//			aPubSub.wasPublishedOnTopic('/politics', true);
	*/	
	aPubSub.prototype.wasPublishedOnTopic = function(myTopic, getBubbles) {
		if (typeof getBubbles != "boolean") getBubbles = false;
		var matchingPublications = new Array();
		for (var i in this.publicationLog) {
			var hit = false;
			if (getBubbles) {
				hit = (this.publicationLog[i].topic.indexOf(myTopic) == 0); 
			} else {
				hit = (this.publicationLog[i].topic.indexOf(myTopic) == myTopic); 
			}
			if (hit) {
				matchingPublications.push(this.publicationLog[i]);
			}
		}
		return matchingPublications;
	}

	/* --------------------------------------------------------------- aPubSub.setDefaults  ----------------------------------------------------------
	// description: 
	//		change the default settings of this aPubSub instance.
	// newDefaults: Object
	//		currently directly supported attributes are:
	//			scope: (object) myScope 
	// 						defaults to myScope, if you want a different scope in the callbacks, just supply it here
	//			persist: (boolean) false,   
	// 						defaults to false, if true the publication will be saved in the this.publicationLog object (with a timestamp attached)
	//			bubble: (boolean) false	
	//  					defaults to false, if true a topic will bubble up it's branch: 
	//							/politics/europe/internet/myOpinion  would activate callbacks for 
	//								first '/politics/europe/internet/myOpinion',
	//								then '/politics/europe/internet',
	//								then '/politics/europe', 
	//								then '/politics',
	//								and '/'
	//		}
	//	returns: Object
	//		{ 
	//			scope: objec t,
	//			persist: boolean,
	//			bubble: boolean 
	//		}
	//	example:
	//			aPubSub.setDefaults({ scope: this, persist: true, bubble: false });
	*/	
	aPubSub.prototype.setDefaults = function(newDefaults) {
		utils.extend(this.defaults, newDefaults);
		return this.defaults;
	}

	/* --------------------------------------------------------------- aPubSub.stripSlashes  ----------------------------------------------------------
	// description: 
	//		used internally. a simply utility function to strip slashes from both ends of the topic string....
	// topic: String
	//
	//	returns: String
	//		
	//
	//	example:
	//			aPubSub.stripSlashes('/news/cnn/'); => 'news/cnn'
	*/
	aPubSub.prototype.stripSlashes = function(topic) {		
		if (typeof topic != 'string') return '';
		var stripped = topic.replace(new RegExp('^[/]+', 'g'), '');
		stripped = topic.replace(new RegExp('[/]+$', 'g'), '');
		return stripped;
	}
	

